# Incursion

TAKEOVERTYPE_CAPTURE = 1
TAKEOVERTYPE_NEUTRALIZE = 2

SCORE_CAPTURE = 4
SCORE_NEUTRALIZE = 2
SCORE_CAPTUREASSIST = 2
SCORE_NEUTRALIZEASSIST = 1
SCORE_DEFEND = 2

Top = 0
Middle = 1
Bottom = 2

class INC_Data:
	def __init__(self):
		self.cp = None
		self.cpTimer = None
		
INC = INC_Data()

import host
import bf2
import math
import tokenize

from game.scoringCommon import addScore, RPL
from game.content import getVehicleType
from game.imports.artillery import onEnterArtyVehicle, onExitArtyVehicle
from game.imports.dod import onDoDCPTrigger, onPlayerDeathDoD, onPlayerKilledDoD, onPlayerSpawnDoD
from game.imports.demoRecording import demoChat
from game.poeSettings import GetSetting
from bf2 import g_debug

g_controlPoints = [] # cache, as this map won't change


usingLauncher = None
try:
	usingLauncher = GetSetting("poe2Launcher")
	usingLauncher = int(usingLauncher[0])
except:
	pass


DOME_OF_DEATH = None
NUKE_TIME_LIMIT = None
DEMO_RECORDING = None
INC_TIME_LIMIT = None
INC_WIN_TIME_LIMIT = None
INC_ROUNDS_PER_MAP = None
INC_FREECAM = None

def handle_token(type, token, (srow, scol), (erow, ecol), line):
	global DOME_OF_DEATH
	global NUKE_TIME_LIMIT
	global DEMO_RECORDING
	global INC_TIME_LIMIT
	global INC_WIN_TIME_LIMIT
	global INC_ROUNDS_PER_MAP
	global INC_FREECAM
	if srow % 2 == 1:
		pass
	else:
		if DOME_OF_DEATH is None:	
			try: DOME_OF_DEATH = int(token)
			except: pass
		elif NUKE_TIME_LIMIT is None:
			try: NUKE_TIME_LIMIT = int(token)
			except: pass
		elif DEMO_RECORDING is None:
			try: DEMO_RECORDING = int(token)
			except: pass
		elif INC_TIME_LIMIT is None:
			try: INC_TIME_LIMIT = int(token)
			except: pass
		elif INC_WIN_TIME_LIMIT is None:
			try: INC_WIN_TIME_LIMIT = int(token)
			except: pass
		elif INC_ROUNDS_PER_MAP is None:
			try: INC_ROUNDS_PER_MAP = int(token)
			except: pass
		elif INC_FREECAM is None:
			try: INC_FREECAM = int(token)
			except: pass

def initSettings():
	global DOME_OF_DEATH
	global NUKE_TIME_LIMIT
	global DEMO_RECORDING
	global INC_TIME_LIMIT
	global INC_WIN_TIME_LIMIT
	global INC_ROUNDS_PER_MAP
	global INC_FREECAM
	if usingLauncher:
		try:
			DOME_OF_DEATH = GetSetting("domeOfDeath")
			DOME_OF_DEATH = int(DOME_OF_DEATH[0])
		
			DEMO_RECORDING = GetSetting("demoRecording")
			DEMO_RECORDING = int(DEMO_RECORDING[0])
			
			INC_TIME_LIMIT = GetSetting("kingOfTheCastleTimeLimit")
			INC_TIME_LIMIT = int(INC_TIME_LIMIT[0])
			
			INC_WIN_TIME_LIMIT = GetSetting("kingOfTheCastleHoldTime")
			INC_WIN_TIME_LIMIT = int(INC_WIN_TIME_LIMIT[0])
		except:
			if g_debug: print 'Error in returning game mode settings'
	else:
		if g_debug: print 'not using laucher'
		try:
			settingsList = open("mods/poe2/settings/poe2settings.con", "r")
			tokenize.tokenize(settingsList.readline, handle_token)
			settingsList.close()
			
		except IOError:
			host.rcon_invoke('echo "WARNING! poe2settings.con not found!"')
			DOME_OF_DEATH = 0
			NUKE_TIME_LIMIT = 0
			DEMO_RECORDING = 0
			INC_TIME_LIMIT = 5
			INC_WIN_TIME_LIMIT = 30
			INC_ROUNDS_PER_MAP = 10
			INC_FREECAM = 1

initSettings()

ORIGINAL_SERVER_TIMELIMIT = None
ORIGINAL_SERVER_ROUNDS_PER_MAP = None
ORIGINAL_SERVER_FREECAM = None

def init():	
	global ORIGINAL_SERVER_TIMELIMIT
	global ORIGINAL_SERVER_ROUNDS_PER_MAP
	global ORIGINAL_SERVER_FREECAM
	global INC_ROUNDS_PER_MAP
	global INC_FREECAM
	global INC_TIME_LIMIT
	global INC_WIN_TIME_LIMIT
	
	# Boundary check on win time
	if INC_WIN_TIME_LIMIT < 0: INC_WIN_TIME_LIMIT = 0
	elif INC_WIN_TIME_LIMIT > 120: INC_WIN_TIME_LIMIT = 120
	elif INC_WIN_TIME_LIMIT % 10 != 0:
		INC_WIN_TIME_LIMIT = math.floor(INC_WIN_TIME_LIMIT)
	
	# round time limit check
	if INC_ROUNDS_PER_MAP < 1: INC_ROUNDS_PER_MAP = 1
	
	#get the settings of the server
	ORIGINAL_SERVER_TIMELIMIT = host.rcon_invoke("sv.timeLimit")
	ORIGINAL_SERVER_ROUNDS_PER_MAP = host.rcon_invoke("sv.roundsPerMap")
	ORIGINAL_SERVER_FREECAM = host.rcon_invoke("sv.allowFreeCam")
	
	#set specific timelimit for king of castle mode
	host.rcon_invoke("sv.timeLimit " + str(INC_TIME_LIMIT))
	#enable freecam
	if INC_FREECAM is not None:
		host.rcon_invoke("sv.allowFreeCam " + str(INC_FREECAM))
	#set rounds
	if INC_ROUNDS_PER_MAP is not None:
		host.rcon_invoke("sv.roundsPerMap " + str(INC_ROUNDS_PER_MAP))
	
	# events hook
	host.registerGameStatusHandler(onGameStatusChangedINC)
	host.sh_setEnableCommander(0)
		
	host.registerHandler('TimeLimitReached', onTimeLimitReachedINC, 1)
	host.registerHandler('PlayerConnect', onPlayerConnectINC, 1)
	
		
def deinit():	
	global ORIGINAL_SERVER_TIMELIMIT
	global ORIGINAL_SERVER_ROUNDS_PER_MAP
	global ORIGINAL_SERVER_FREECAM
	#set back the original settings of the server
	host.rcon_invoke("sv.timeLimit " + ORIGINAL_SERVER_TIMELIMIT)
	host.rcon_invoke("sv.roundsPerMap " + ORIGINAL_SERVER_ROUNDS_PER_MAP)
	host.rcon_invoke("sv.allowFreeCam " + ORIGINAL_SERVER_FREECAM)
	
	bf2.triggerManager.destroyAllTriggers()
	global g_controlPoints
	g_controlPoints = []
	host.unregisterGameStatusHandler(onGameStatusChangedINC)

def onGameStatusChangedINC(status):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	global g_controlPoints
	global incDefendingTeam
	global incAttackingTeam
	global incRespawnTime
	global incRespawnTimer
	global incRoundStarted
	
	if g_debug: print 'Incursion game mode playing'
	if status == bf2.GameStatus.Playing:
		# add control point triggers
		g_controlPoints = bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint')
		for obj in g_controlPoints:
			if DOME_OF_DEATH:
				if obj.cp_getParam('unableToChangeTeam') == 1:
					radius = float(obj.getTemplateProperty('radius'))
					id = bf2.triggerManager.createRadiusTrigger(obj, onDoDCPTrigger, '<<PCO>>', radius, (1, 2, 3))			
					obj.triggerId = id
					obj.lastAttackingTeam = 0
					if obj.cp_getParam('team') > 0:
						obj.flagPosition = Top
					else:
						obj.flagPosition = Bottom
				else:
					radius = float(obj.getTemplateProperty('radius'))
					isHemi = int(obj.cp_getParam('isHemisphere'))
					if isHemi != 0:
						id = bf2.triggerManager.createHemiSphericalTrigger(obj, onINCTrigger, '<<PCO>>', radius, (1, 2, 3))
					else:
						id = bf2.triggerManager.createRadiusTrigger(obj, onINCTrigger, '<<PCO>>', radius, (1, 2, 3))			
					obj.triggerId = id
					obj.lastAttackingTeam = 0
					team = obj.cp_getParam('team')
					if team > 0:
						obj.flagPosition = Top
						incDefendingTeam = team
						
						if team == 1:
							incAttackingTeam = 2
						elif team == 2:
							incAttackingTeam = 1
					else:
						obj.flagPosition = Bottom
			else:
				radius = float(obj.getTemplateProperty('radius'))
				isHemi = int(obj.cp_getParam('isHemisphere'))
				if isHemi != 0:
					id = bf2.triggerManager.createHemiSphericalTrigger(obj, onINCTrigger, '<<PCO>>', radius, (1, 2, 3))
				else:
					id = bf2.triggerManager.createRadiusTrigger(obj, onINCTrigger, '<<PCO>>', radius, (1, 2, 3))			
				obj.triggerId = id
				obj.lastAttackingTeam = 0
				team = obj.cp_getParam('team')
				if team > 0:
					obj.flagPosition = Top
					defendingTeam = team
					
					if team == 1:
						incAttackingTeam = 2
					elif team == 2:
						incAttackingTeam = 1
				else:
					obj.flagPosition = Bottom


		host.registerHandler('ControlPointChangedOwner', onCPStatusChangeINC)

		ticketsTeam1 = calcStartTickets(bf2.gameLogic.getDefaultTickets(1))
		ticketsTeam2 = calcStartTickets(bf2.gameLogic.getDefaultTickets(2))
		
		bf2.gameLogic.setTickets(1, ticketsTeam1)
		bf2.gameLogic.setTickets(2, ticketsTeam2)

		bf2.gameLogic.setTicketState(1, 0)
		bf2.gameLogic.setTicketState(2, 0)

		bf2.gameLogic.setTicketLimit(1, 1, 0)
		bf2.gameLogic.setTicketLimit(2, 1, 0)
		bf2.gameLogic.setTicketLimit(1, 2, 10)
		bf2.gameLogic.setTicketLimit(2, 2, 10)
		bf2.gameLogic.setTicketLimit(1, 3, int(ticketsTeam1*0.1))
		bf2.gameLogic.setTicketLimit(2, 3, int(ticketsTeam2*0.1))
		bf2.gameLogic.setTicketLimit(1, 4, int(ticketsTeam1*0.2))
		bf2.gameLogic.setTicketLimit(2, 4, int(ticketsTeam1*0.2))
		
		host.registerHandler('TicketLimitReached', onTicketLimitReachedINC)
		updateTicketLoss()


		# player events
		host.registerHandler('PlayerDeath', onPlayerDeathINC)
		host.registerHandler('PlayerKilled', onPlayerKilledINC)
		host.registerHandler('PlayerRevived', onPlayerRevivedINC)
		host.registerHandler('PlayerSpawn', onPlayerSpawnINC)
		host.registerHandler('EnterVehicle', onEnterVehicleINC)
		host.registerHandler('ExitVehicle', onExitVehicleINC)
		if DEMO_RECORDING: host.registerHandler('ChatMessage', onChatMessageINC)
		
		#Set respawn time
		startDelay = host.rcon_invoke("sv.startDelay")
		incRespawnTime = int(bf2.serverSettings.getTimeLimit()) + int(startDelay)
		#start timer to update the respawn time
		if incRespawnTimer != None:
			incRespawnTimer.destroy()
			incRespawnTimer = None
		
		incRespawnTimer = bf2.Timer(updateRespawnTime, 1, 1)
		incRespawnTimer.setRecurring(1)
		
		# round has not started yet		
		incRoundStarted = 0

	else:
		bf2.triggerManager.destroyAllTriggers()
		g_controlPoints = []
		
def onChatMessageINC(playerId, text, channel, flags):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	player = bf2.playerManager.getPlayerByIndex(playerId)
	demoChat(player, text)

def calcStartTickets(mapDefaultTickets):
	return int(mapDefaultTickets * (bf2.serverSettings.getTicketRatio() / 50.0))
	
def onTimeLimitReachedINC(value):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	#Defending team win when time runs out
	global incDefendingTeam
	
	# zero out tickets for losing team
	if (incDefendingTeam == 1):
		bf2.gameLogic.setTickets(2, 0)
	elif (incDefendingTeam == 2):
		bf2.gameLogic.setTickets(1, 0)
	host.sgl_endGame(incDefendingTeam, 3)

# update ticket system
def updateTicketLoss():
	areaValueTeam1 = 0
	areaValueTeam2 = 0
	totalAreaValue = 0
	numCpsTeam0 = 0
	numCpsTeam1 = 0
	numCpsTeam2 = 0
	
	# calculate control point area value for each team
	for obj in g_controlPoints:
		team = obj.cp_getParam('team')
		if team == 1:
			areaValueTeam1 += obj.cp_getParam('areaValue', team)
			totalAreaValue += areaValueTeam1
			numCpsTeam1 += 1
		elif team == 2:
			areaValueTeam2 += obj.cp_getParam('areaValue', team)
			totalAreaValue += areaValueTeam2
			numCpsTeam2 += 1
		else:
			numCpsTeam0 += 1
			totalAreaValue += 0
	
	# check if a team has no control points
	if numCpsTeam1 == 0 or numCpsTeam2 == 0:
		if numCpsTeam1 == 0:
			losingTeam = 1
			winningTeam = 2
		else:
			losingTeam = 2
			winningTeam = 1
		
		# check if there is anyone alive
		foundLivingPlayer = False
		players = bf2.playerManager.getPlayers()
		for p in players:
			if p.getTeam() == losingTeam and p.isAlive():
				foundLivingPlayer = True
				break
				
		if not foundLivingPlayer:
			ticketLossPerSecond = bf2.gameLogic.getDefaultTicketLossAtEndPerMin()
			bf2.gameLogic.setTicketChangePerSecond(losingTeam, -ticketLossPerSecond)
			bf2.gameLogic.setTicketChangePerSecond(winningTeam, 0)

			return
	
	# update ticket loss
	team1AreaOverweight = areaValueTeam1 - areaValueTeam2
	percentualOverweight = 1.0
	if totalAreaValue != 0:
		percentualOverweight = abs(team1AreaOverweight / totalAreaValue)
	
	ticketLossPerSecTeam1 = calcTicketLossForTeam(1, areaValueTeam2, -team1AreaOverweight)
	ticketLossPerSecTeam2 = calcTicketLossForTeam(2, areaValueTeam1, team1AreaOverweight)
	bf2.gameLogic.setTicketChangePerSecond(1, -ticketLossPerSecTeam1)
	bf2.gameLogic.setTicketChangePerSecond(2, -ticketLossPerSecTeam2)

	
# actual ticket loss calculation function
def calcTicketLossForTeam(team, otherTeamAreaValue, otherTeamAreaOverweight):
	if otherTeamAreaValue >= 100 and otherTeamAreaOverweight > 0:
		ticketLossPerSecond = (bf2.gameLogic.getDefaultTicketLossPerMin(team) / 60.0) * (otherTeamAreaOverweight / 100.0)
		return ticketLossPerSecond
	else:
		return 0
	
DOWNWARDS = -1
UPWARDS = 1	

# called when tickets reach a predetermined limit (negativ value means that the tickets have become less than the limit)
def onTicketLimitReachedINC(team, limitId):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	if (limitId == -1):
		if (team == 1):
			winner = 2
		
		elif (team == 2):
			winner = 1
		
		bf2.gameLogic.setTicketState(1, 0)
		bf2.gameLogic.setTicketState(2, 0)
	
		host.sgl_endGame(winner, 3)		
	else:
		updateTicketWarning(team, limitId)


# called when the ticket state should be updated (for triggering messages and sounds based on tickets left)
def updateTicketWarning(team, limitId):
	oldTicketState = bf2.gameLogic.getTicketState(team)
	newTicketState = 0
	
	if (oldTicketState >= 10):
		newTicketState = 10		

	if (limitId == -2):
		newTicketState = 10
	
	elif (limitId == 2):
		newTicketState = 0		

	elif (limitId == -3):
		newTicketState += 2

	elif (limitId == -4):
		newTicketState += 1

	if (oldTicketState != newTicketState):
		bf2.gameLogic.setTicketState(team, newTicketState)
		
	
# called when someone enters or exits cp radius
def onINCTrigger(triggerId, cp, vehicle, enter, userData):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	global INC
	global incDefendingTeam
	
	if not cp.isValid(): return	
	
	playersInVehicle = None	
	if vehicle:
		playersInVehicle = vehicle.getOccupyingPlayers()
	
	if enter:
		for p in playersInVehicle:
			cp = getOccupyingCP(p)
			if cp != None:
				if not p.getIsInsideCP():
					p.enterCpAt = host.timer_getWallTime()
					
	if vehicle:	
		for p in playersInVehicle:
				if p.getVehicle() == p.getDefaultVehicle():
					p.setIsInsideCP(enter)


	# count people in radius
	team1Occupants = 0
	team2Occupants = 0

	pcos = bf2.triggerManager.getObjects(cp.triggerId)
	for o in pcos:
		if not o: continue # dep: you can get None in the result tuple when the host can't figure out what object left the trigger
		if o.getParent(): continue # jlo: getOccupyingPlayers returns all players downwards in the hierarchy, so dont count them twice
		occupyingPlayers = o.getOccupyingPlayers()
		for p in occupyingPlayers:
			if p.isAlive() and not p.isManDown() and not p.killed and p.getVehicle() == p.getDefaultVehicle():
				pTeam = p.getTeam()
				if pTeam == 1:
					team1Occupants += 1
				else:
					team2Occupants += 1


	# determine who is actually getting it
	team1OverWeight = team1Occupants - team2Occupants
	attackOverWeight = 0

	if team1OverWeight > 0:
		attackingTeam = 1
	elif team1OverWeight < 0:
		attackingTeam = 2
	else:
		attackingTeam = 0

	
	if team1Occupants == 0 and team2Occupants == 0:

		# nobody here, slowly go back to owning team
		if cp.cp_getParam('team') == 0:
			attackOverWeight = -0.5
		else:
			attackOverWeight = 0.5
			
		timeToChangeControl = cp.cp_getParam('timeToLoseControl')

	else:

		# raise flag if already ours, or at bottom and neutral. Otherwise lower first.
		if cp.cp_getParam('flag') == attackingTeam or (cp.flagPosition == Bottom and cp.cp_getParam('team') == 0):

			# our flag, raise
			attackOverWeight = abs(team1OverWeight)
			timeToChangeControl = cp.cp_getParam('timeToGetControl')
		else:

			# other team raised flag, lower first
			attackOverWeight = - abs(team1OverWeight)
			timeToChangeControl = cp.cp_getParam('timeToLoseControl')


	if cp.cp_getParam('onlyTakeableByTeam') != 0 and cp.cp_getParam('onlyTakeableByTeam') != attackingTeam:
		return
	
	if cp.cp_getParam('team') == incDefendingTeam and INC.cp != None and attackingTeam != incDefendingTeam:
		return


	# flag can only be changed when at bottom
	if cp.flagPosition == Bottom:
		cp.cp_setParam('flag', attackingTeam)


	# calculate flag raising/lowering speed
	if timeToChangeControl > 0:
		takeOverChangePerSecond = 1.0 * attackOverWeight / timeToChangeControl
	else:
		takeOverChangePerSecond = 0.0

	if (cp.flagPosition == Top and takeOverChangePerSecond > 0) or (cp.flagPosition == Bottom and takeOverChangePerSecond < 0):
		takeOverChangePerSecond = 0.0

	if abs(takeOverChangePerSecond) > 0:
		cp.flagPosition = Middle
				
	cp.cp_setParam('takeOverChangePerSecond', takeOverChangePerSecond)


# called when a control point flag reached top or bottom
def onCPStatusChangeINC(cp, top):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	global winTime
	global INC_WIN_TIME_LIMIT
	global incDefendingTeam
	global incAttackingTeam
	global INC
	
	playerId = -1
	takeoverType = -1
	newTeam = -1
	scoringTeam = -1
	
	if top:	cp.flagPosition = Top
	else:   cp.flagPosition = Bottom
	
	# determine capture / neutralize / defend
	if cp.cp_getParam('team') != 0:
		if top:
			pass
		else:
			if cp.cp_getParam('team') == incDefendingTeam and INC.cp != None:
				#host.rcon_invoke('game.sayTeam ' + str(incAttackingTeam) + " \"*** YOU CAN ONLY HOLD ONE CONTROLPOINT AT A TIME ***\"")
				players = cp.getOccupyingPlayers()
				sendPlayersMessage(players, 1031619, 3)
			
			else:	
				# neutralize
				newTeam = 0
				if cp.cp_getParam('team') == 1:
					scoringTeam = 2
				else:
					scoringTeam = 1
					
				takeoverType = TAKEOVERTYPE_NEUTRALIZE
	else:
		if top:
			# capture
			newTeam = cp.cp_getParam('flag')
			scoringTeam = newTeam
			takeoverType = TAKEOVERTYPE_CAPTURE
		else:
			pass
			# hit bottom, but still neutral

	# scoring
	if takeoverType > 0:
		pcos = bf2.triggerManager.getObjects(cp.triggerId)
	
		# count number of players
		scoringPlayers = []
		firstPlayers = []
		for o in pcos:
			occupyingPlayers = o.getOccupyingPlayers()
			if not occupyingPlayers:
				pass 
			
			for p in occupyingPlayers:
				playerVehicle = p.getVehicle()
				parent = playerVehicle.getParent()
				# do not count player in helicopter
				if not parent:
					if getVehicleType(playerVehicle.templateName) == 3:
						continue
				else:
					if getVehicleType(parent.templateName) == 3:
						continue

				if p.isAlive() and not p.isManDown() and p.getTeam() == scoringTeam:
					if len(firstPlayers) == 0 or p.enterCpAt < firstPlayers[0].enterCpAt:
						firstPlayers = [p]
					elif p.enterCpAt == firstPlayers[0].enterCpAt:
						firstPlayers += [p]
					
					if not p in scoringPlayers:
						scoringPlayers += [p]
	
		# deal score
		for p in scoringPlayers:
			oldScore = p.score.score;
			if takeoverType == TAKEOVERTYPE_CAPTURE:				
				if p in firstPlayers:
					p.score.cpCaptures += 1
					addScore(p, SCORE_CAPTURE, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 0) #12 = Conquest, 0 = Capture
					playerId = p.index
					
				else:
					p.score.cpAssists += 1
					addScore(p, SCORE_CAPTUREASSIST, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 2) #12 = Conquest, 2 = Assist


			elif takeoverType == TAKEOVERTYPE_NEUTRALIZE:					
				if p in firstPlayers:
					p.score.cpNeutralizes += 1
					addScore(p, SCORE_NEUTRALIZE, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 3) #12 = Conquest, 3 = Neutralize
				else:
					p.score.cpNeutralizeAssists += 1
					addScore(p, SCORE_NEUTRALIZEASSIST, RPL)
					bf2.gameLogic.sendGameEvent(p, 12, 4) #12 = Conquest, 4 = Neutralize assist
		
		#Set INC timer
		if takeoverType == TAKEOVERTYPE_CAPTURE:
			if INC.cp is None:
				if newTeam != incDefendingTeam:
					INC.cp = cp
					winTime = INC_WIN_TIME_LIMIT
					INC.cpTimer = bf2.Timer(endINCRound, 1, 1)
					INC.cpTimer.setRecurring(1)
					#host.rcon_invoke('game.sayTeam ' + str(incAttackingTeam) + " \"*** HOLD THE CONTROLPOINT FOR " + str(INC_WIN_TIME_LIMIT) + " SECONDS ***\"")
					#host.rcon_invoke('game.sayTeam ' + str(incDefendingTeam) + " \"*** GET THE CONTROLPOINT BACK WITHIN " + str(INC_WIN_TIME_LIMIT) + " SECONDS ***\"")
					sendCPCapturedMessage(incAttackingTeam, incDefendingTeam)
		
		elif takeoverType == TAKEOVERTYPE_NEUTRALIZE:
			if INC.cpTimer != None and INC.cp == cp:
				INC.cp = None
				INC.cpTimer.destroy()
				INC.cpTimer = None
				#host.rcon_invoke('game.sayTeam ' + str(incAttackingTeam) + " \"*** YOU LOST THE CONTROLPOINT ***\"")
				sendTeamMessage(incAttackingTeam, 1031119, 2)
				#host.rcon_invoke('game.sayTeam ' + str(incDefendingTeam) + " \"*** YOU REGAINED THE CONTROLPOINT ***\"")
				sendTeamMessage(incDefendingTeam, 1031119, 3)

	# immediate ticket loss for opposite team
	enemyTicketLossInstant = cp.cp_getParam('enemyTicketLossWhenCaptured')
	if enemyTicketLossInstant > 0 and newTeam > 0:
		
		if newTeam == 1:
			punishedTeam = 2
		elif newTeam == 2:
			punishedTeam = 1
		
		tickets = bf2.gameLogic.getTickets(punishedTeam)
		tickets -= enemyTicketLossInstant
		bf2.gameLogic.setTickets(punishedTeam, tickets)
	
	
	# update control point	
	cp.cp_setParam('playerId', playerId) #always set player first
	if newTeam != -1 and cp.cp_getParam('team') != newTeam:
		cp.cp_setParam('team', newTeam)
	onINCTrigger(cp.triggerId, cp, 0, 0, 0)
	updateTicketLoss()

def endINCRound(data):
	global winTime
	global INC
	
	winTime -= 1
	
	#print "winTime: ", winTime
	
	if (winTime <= 0):
		winningTeam = INC.cp.cp_getParam('team')
		
		# zero out tickets for losing team
		if (winningTeam == 1):
			bf2.gameLogic.setTickets(2, 0)
		elif (winningTeam == 2):
			bf2.gameLogic.setTickets(1, 0)
		
		INC.cp = None
		if INC.cpTimer != None:
			INC.cpTimer.destroy()
			INC.cpTimer = None
		
		host.sgl_endGame(winningTeam, 3)
	
	#elif ((winTime % 10) == 0):
	#	#print "(winTime % 10) == 0"
	#	host.rcon_invoke('game.sayAll \"*** ' + str(winTime) + " SECONDS REMAINING! ***"'"')
	
	elif winTime == 60:
		sendAllMessage(1190507, 3)
	elif winTime == 45:
		sendAllMessage(1191819, 1)
	elif winTime == 30:
		sendAllMessage(1191819, 2)
	elif winTime == 15:
		sendAllMessage(1191819, 3)
	elif winTime == 10:
		sendAllMessage(1190304, 1)
	elif winTime == 5:
		sendAllMessage(1190304, 2)
		
	
def endINCRoundNoPlayers(team):
	global incDefendingTeam
	# zero out tickets for losing team
	winningTeam = incDefendingTeam
	if (team == 1):
		bf2.gameLogic.setTickets(1, 0)
		winningTeam = 2
	elif (team == 2):
		bf2.gameLogic.setTickets(2, 0)
		winningTeam = 1
		
	host.sgl_endGame(winningTeam, 3)
	
				
def onPlayerDeathINC(victim, vehicle):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	if not victim: return
	
	global incRespawnTime
	global incRoundStarted
	
	# destroy DoD timer
	if DOME_OF_DEATH:
		onPlayerDeathDoD(victim)

	# punish team with one ticket
	if victim != None:
		team = victim.getTeam()
		teamTickets = bf2.gameLogic.getTickets(team)
		teamTickets -= 1
		bf2.gameLogic.setTickets(team, teamTickets)
		

	# check if it was the last player
	foundLivingPlayer = False
	players = bf2.playerManager.getPlayers()
	for p in players:
		if p != victim and p.getTeam() == victim.getTeam() and p.isAlive():
			foundLivingPlayer = True
	
	# end game if this was last player alive for this team
	if not foundLivingPlayer:
		#updateTicketLoss()
		endINCRoundNoPlayers(victim.getTeam())
	
	#change respawn time for player
	victim.setTimeToSpawn(incRespawnTime)
	if incRoundStarted == 0:
		incRoundStarted = 1



def onPlayerKilledINC(victim, attacker, weapon, assists, object):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	if not victim: return
	
	if DOME_OF_DEATH:
		onPlayerKilledDoD(victim, attacker)

	victim.killed = True # Remove this flag to enabled players in "man down"-state to captuer flags
	
	# update flag takeover status if victim was in a CP radius
	cp = getOccupyingCP(victim)
	if cp != None:
		onINCTrigger(-1, cp, victim.getVehicle(), False, None)

		# give defend score if killing enemy within cp radius
		if attacker != None and attacker.getTeam() != victim.getTeam()\
		   and cp.cp_getParam('unableToChangeTeam') == 0 and cp.cp_getParam('onlyTakeableByTeam') == 0:
		
			##if g_debug: print "CP found:", cp.templateName
			if cp != None and cp.cp_getParam('team') == attacker.getTeam():
				#if g_debug: print "Player %d defended %s" % (attacker.index, cp.templateName)
				attacker.score.cpDefends += 1
				addScore(attacker, SCORE_DEFEND, RPL)
				bf2.gameLogic.sendGameEvent(attacker, 12, 1) #12 = Conquest, 1 = Defend


def onPlayerRevivedINC(victim, attacker):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	# update flag takeover status if victim was in a CP radius
	victim.killed = False
	
	cp = getOccupyingCP(victim)
	if cp != None:
		if DOME_OF_DEATH and cp.cp_getParam('unableToChangeTeam') != 0:
			onDoDCPTrigger(-1, cp, victim.getVehicle(), True, None)
		else:
			onINCTrigger(-1, cp, victim.getVehicle(), True, None)
	
def onPlayerSpawnINC(player, soldier):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	player.killed = False
	
	if DOME_OF_DEATH:
		onPlayerSpawnDoD(player)
	
	if player.firstSpawn:
		if not DOME_OF_DEATH:
			bf2.gameLogic.sendMedalEvent(player, 1031406, 3) # dod is off
		
		global INC_WIN_TIME_LIMIT
		
		# Send info about time to hold CP to win/lose
	
		if INC_WIN_TIME_LIMIT == 0:
			bf2.gameLogic.sendMedalEvent(player, 1190304, 3)
		elif INC_WIN_TIME_LIMIT == 10:
			bf2.gameLogic.sendMedalEvent(player, 1220118, 1)
		elif INC_WIN_TIME_LIMIT == 20:
			bf2.gameLogic.sendMedalEvent(player, 1220118, 2)
		elif INC_WIN_TIME_LIMIT == 30:
			bf2.gameLogic.sendMedalEvent(player, 1220118, 3)
		elif INC_WIN_TIME_LIMIT == 40:
			bf2.gameLogic.sendMedalEvent(player, 1222016, 1)
		elif INC_WIN_TIME_LIMIT == 50:
			bf2.gameLogic.sendMedalEvent(player, 1222016, 2)
		elif INC_WIN_TIME_LIMIT == 60:
			bf2.gameLogic.sendMedalEvent(player, 1222016, 3)
		elif INC_WIN_TIME_LIMIT == 70:
			bf2.gameLogic.sendMedalEvent(player, 1220803, 1)
		elif INC_WIN_TIME_LIMIT == 80:
			bf2.gameLogic.sendMedalEvent(player, 1220803, 2)
		elif INC_WIN_TIME_LIMIT == 90:
			bf2.gameLogic.sendMedalEvent(player, 1220803, 3)
		elif INC_WIN_TIME_LIMIT == 100:
			bf2.gameLogic.sendMedalEvent(player, 1220122, 1)
		elif INC_WIN_TIME_LIMIT == 110:
			bf2.gameLogic.sendMedalEvent(player, 1220122, 2)
		elif INC_WIN_TIME_LIMIT == 120:
			bf2.gameLogic.sendMedalEvent(player, 1220122, 3)
		
		player.firstSpawn = False

	
def onEnterVehicleINC(player, vehicle, freeSoldier = False):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	# update flag takeover status if player in a CP radius
	player.setIsInsideCP(False)
	cp = getOccupyingCP(player)
	if cp != None:
		if DOME_OF_DEATH and cp.cp_getParam('unableToChangeTeam') == 1:
			pass #onDoDCPTrigger(-1, cp, vehicle, False, None)
		else:
			onCPTrigger(-1, cp, vehicle, False, None)
		updateCaptureStatus(vehicle) 

def onExitVehicleINC(player, vehicle):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	# update flag takeover status if player in a CP radius
	cp = getOccupyingCP(player)
	# can this cp be captured at all?
	player.setIsInsideCP(cp != None and cp.cp_getParam('unableToChangeTeam') == 0)
	updateCaptureStatus(vehicle)

#Update cp capture status on players in vehicle
def updateCaptureStatus(vehicle):	
	rootVehicle = bf2.objectManager.getRootParent(vehicle)
	playersInVehicle = rootVehicle.getOccupyingPlayers()
	vehicleType = getVehicleType(rootVehicle.templateName)
	
	# set players in helicopters as outside, others inside
	for p in playersInVehicle:
		cp = getOccupyingCP(p)
		p.setIsInsideCP(cp != None and cp.cp_getParam('unableToChangeTeam') == 0 and vehicleType != 3)
		
# find cp that player is occupying, if any		
def getOccupyingCP(player):
	vehicle = player.getVehicle()
	playerPos = vehicle.getPosition()
	
	# find closest CP
	closestCP = None
	
	if len(g_controlPoints) == 0: return None
	
	for obj in g_controlPoints:
		distanceTo = getVectorDistance(playerPos, obj.getPosition())
		if closestCP == None or distanceTo < closestCPdist:
			closestCP = obj
			closestCPdist = distanceTo
	
	# is the player in radius?
	pcos = bf2.triggerManager.getObjects(closestCP.triggerId)
	for o in pcos:
		if o == player.getDefaultVehicle():
			# Player is DEFAULT vehicle - this is needed when called from onEnterVehicle
			return closestCP
		else:
			occupyingPlayers = o.getOccupyingPlayers()
			for p in occupyingPlayers:
				if p == player:
					return closestCP
	return None

# get distance between two positions
def getVectorDistance(pos1, pos2):
	diffVec = [0.0, 0.0, 0.0]
	diffVec[0] = math.fabs(pos1[0] - pos2[0])
	diffVec[1] = math.fabs(pos1[1] - pos2[1])
	diffVec[2] = math.fabs(pos1[2] - pos2[2])
	 
	return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])


#update the respawn time
def updateRespawnTime(data):
	global incRespawnTime
	incRespawnTime -= 1

def onPlayerConnectINC(player):
	if bf2.serverSettings.getGameMode() != "gpm_inc": return
	
	player.firstSpawn = True
	
	global incRoundStarted
	global incRespawnTime
	
	if incRoundStarted == 1:
		player.setTimeToSpawn(incRespawnTime)


# Send specific message (id + data) to an entire team
def sendTeamMessage(team, id, data):
	players = bf2.playerManager.getPlayers()
	for p in players:
		if p.getTeam() == team and p.isAlive():
			bf2.gameLogic.sendMedalEvent(p, id, data)

def sendAllMessage(id, data):
	players = bf2.playerManager.getPlayers()
	for p in players:
		if p.isAlive():
			bf2.gameLogic.sendMedalEvent(p, id, data)

def sendPlayersMessage(players, id, data):
	for p in players:
		if p.isAlive():
			bf2.gameLogic.sendMedalEvent(p, id, data)

# Pick message to get from value (time)
def sendCPCapturedMessage(attackTeam, defendTeam):
	global INC_WIN_TIME_LIMIT
	
	if INC_WIN_TIME_LIMIT == 0: return
	elif INC_WIN_TIME_LIMIT == 10:
		sendTeamMessage(attackTeam, 1031120, 2)
		sendTeamMessage(defendTeam, 1031105, 3)
	elif INC_WIN_TIME_LIMIT == 20:
		sendTeamMessage(attackTeam, 1031120, 3)
		sendTeamMessage(defendTeam, 1031113, 1)
	elif INC_WIN_TIME_LIMIT == 30:
		sendTeamMessage(attackTeam, 1031109, 2)
		sendTeamMessage(defendTeam, 1031113, 2)
	elif INC_WIN_TIME_LIMIT == 40:
		sendTeamMessage(attackTeam, 1031109, 3)
		sendTeamMessage(defendTeam, 1031113, 3)
	elif INC_WIN_TIME_LIMIT == 50:
		sendTeamMessage(attackTeam, 1031115, 1)
		sendTeamMessage(defendTeam, 1032415, 1)
	elif INC_WIN_TIME_LIMIT == 60:
		sendTeamMessage(attackTeam, 1031115, 2)
		sendTeamMessage(defendTeam, 1032415, 2)
	elif INC_WIN_TIME_LIMIT == 70:
		sendTeamMessage(attackTeam, 1031115, 3)
		sendTeamMessage(defendTeam, 1032415, 3)
	elif INC_WIN_TIME_LIMIT == 80:
		sendTeamMessage(attackTeam, 1031121, 1)
		sendTeamMessage(defendTeam, 1190601, 1)
	elif INC_WIN_TIME_LIMIT == 90:
		sendTeamMessage(attackTeam, 1031121, 2)
		sendTeamMessage(defendTeam, 1190601, 2)
	elif INC_WIN_TIME_LIMIT == 100:
		sendTeamMessage(attackTeam, 1031121, 3)
		sendTeamMessage(defendTeam, 1190601, 3)
	elif INC_WIN_TIME_LIMIT == 110:
		sendTeamMessage(attackTeam, 1031105, 1)
		sendTeamMessage(defendTeam, 1190507, 1)
	elif INC_WIN_TIME_LIMIT == 120:
		sendTeamMessage(attackTeam, 1031105, 2)
		sendTeamMessage(defendTeam, 1190507, 2)
